/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionEndOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionStartOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.clearNow;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.now;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.setNow;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.deny;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.denyIndefinitely;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.grant;

public final class TestPassportEvaluator implements PassportEvaluator {

    @NonNull private static final String NO_NEGATIVE_TEST_RESULTS = "no negative test results";
    @NonNull private static final String POSITIVE_PCR_TEST_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";
    @NonNull private static final String MOST_RECENT_RESULT_IS_POSITIVE = "most recent result is positive";

    @NonNull private final TemporalAmount minimumPeriod;
    @NonNull private final TemporalAmount maximumPeriod;
    @NonNull private final TemporalAmount suspensionPeriod;
    @NonNull private final boolean keepTimeOnMinimumPeriod;
    @NonNull private final boolean keepTimeOnMaximumPeriod;
    @NonNull private final boolean keepTimeOnSuspensionPeriod;

    public TestPassportEvaluator(
        @NonNull final TemporalAmount minimumPeriod, final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriod, final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount suspensionPeriod, final boolean keepTimeOnSuspensionPeriod) {
        this.minimumPeriod = minimumPeriod;
        this.maximumPeriod = maximumPeriod;
        this.suspensionPeriod = suspensionPeriod;
        this.keepTimeOnMinimumPeriod = keepTimeOnMinimumPeriod;
        this.keepTimeOnMaximumPeriod = keepTimeOnMaximumPeriod;
        this.keepTimeOnSuspensionPeriod = keepTimeOnSuspensionPeriod;
    }

    /**
     * Evaluates lab results in the context of test passport rules.
     * @return the result of the evaluation.
     */
    @NonNull
    @Override
    public final PassportEvaluation evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results, @NonNull final LocalDateTime now) {
        try {
            setNow(now);
            return firstNonNullOrDefaultSupplier(
                () -> denyIndefinitely(NO_NEGATIVE_TEST_RESULTS),
                // Reasons to grant or deny a test passport follow below,
                // if none are met return above
                () -> denyForMostRecentTestIsPositive(results),
                () -> grantForNegativePcrTestWithinValidityPeriod(results),
                () -> grantOrDenyForNegativeAntigenTest(results));
        } finally {
            clearNow();
        }
    }

    @Nullable
    private PassportEvaluation denyForMostRecentTestIsPositive(@NonNull final LabResults results) {
        final LabResult mostRecentResult = results.mostRecentResult();
        if(mostRecentResult == null) return null;
        if(mostRecentResult.negative()) return null;

        return denyIndefinitely(MOST_RECENT_RESULT_IS_POSITIVE);
    }

    @Nullable
    private PassportEvaluation grantForNegativePcrTestWithinValidityPeriod(@NonNull final LabResults results) {
        final LabResult.NegativePcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();

        if(mostRecentNegativePcrResult == null) return null;
        if(negativeResultIsWithinValidityPeriod(mostRecentNegativePcrResult)) return grant(validUntil(mostRecentNegativePcrResult));

        return null;
    }

    @Nullable
    private PassportEvaluation grantOrDenyForNegativeAntigenTest(@NonNull final LabResults results) {
        final LabResult.NegativeAntigenResult mostRecentNegativeAntigenResult = results.mostRecentNegativeAntigenResult();
        final LabResult.PositivePcrResult mostRecentPositivePcrResult = results.mostRecentPositivePcrResult();

        if(mostRecentNegativeAntigenResult == null) return null;
        if(negativeResultIsNotWithinValidityPeriod(mostRecentNegativeAntigenResult)) return null;

        return denyOrGrantForNegativeAntigenResultSuspendedByPositivePcrResult(
            mostRecentNegativeAntigenResult, mostRecentPositivePcrResult);
    }


    @NonNull
    private PassportEvaluation denyOrGrantForNegativeAntigenResultSuspendedByPositivePcrResult(
        @NonNull final LabResult.NegativeAntigenResult mostRecentAntigenResult,
        @Nullable final LabResult.PositivePcrResult mostRecentPositivePcrResult) {

        if(mostRecentPositivePcrResult == null) return grant(validUntil(mostRecentAntigenResult));
        if(positivePcrResultWithinSuspensionPeriod(mostRecentPositivePcrResult)) {
            return deny(
                endOfSuspensionPeriod(mostRecentPositivePcrResult),
                POSITIVE_PCR_TEST_WITHIN_SUSPENSION_PERIOD);
        }

        return grant(validUntil(mostRecentAntigenResult));

    }

    @SuppressWarnings({"java:S1854"})
    private boolean negativeResultIsWithinValidityPeriod(@NonNull final LabResult.NegativeResult negativeLabResult) {
        final LocalDateTime startOfValidityPeriod = resolutionStartOfDay(now().plus(minimumPeriod), keepTimeOnMinimumPeriod);
        final LocalDateTime endOfValidityPeriod = resolutionEndOfDay(now().plus(maximumPeriod), keepTimeOnMaximumPeriod);
        final LocalDateTime resultDateTime = negativeLabResult.resultDateTime();

        return !startOfValidityPeriod.isAfter(resultDateTime) && !resultDateTime.isAfter(endOfValidityPeriod);
    }

    private boolean negativeResultIsNotWithinValidityPeriod(@NonNull final LabResult.NegativeResult negativeLabResult) {
        return ! negativeResultIsWithinValidityPeriod(negativeLabResult);
    }

    private boolean positivePcrResultWithinSuspensionPeriod(@NonNull final PcrResult pcrResult) {
        final LocalDateTime startOfSuspensionPeriod = resolutionStartOfDay(now().minus(suspensionPeriod), keepTimeOnSuspensionPeriod);

        return ! startOfSuspensionPeriod.isBefore(pcrResult.resultDateTime());
    }

    @NonNull
    private LocalDateTime endOfSuspensionPeriod(@NonNull final PcrResult positivePcrResult) {
        return resolutionEndOfDay(positivePcrResult.resultDateTime().plus(suspensionPeriod), keepTimeOnSuspensionPeriod);
    }

    @NonNull
    private LocalDateTime validUntil(@NonNull final LabResult negativeLabResult) {
        return resolutionEndOfDay(negativeLabResult.resultDateTime().plus(maximumPeriod), keepTimeOnMaximumPeriod);
    }
}
