/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportDeniedEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;

import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionEndOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.LocalDateTimes.resolutionStartOfDay;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.now;
import static dk.sundhedsdatastyrelsen.cbs.rules.StableLocalDateTime.setNow;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.deny;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.denyIndefinitely;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation.grant;

public final class VaccinationPassportEvaluator implements PassportEvaluator {

    @NonNull private static final String NOT_COMPLETED_VACCINATION = "vaccination not completed";
    @NonNull private static final String MINIMUM_DURATION_NOT_PASSED = "minimum duration not passed";
    @NonNull private static final String VACCINATION_EXPIRED = "vaccination expired";
    @NonNull private static final String POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT = "postive antigen result within suspension period without subsequent negative pcr result";
    @NonNull private static final String POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";

    @NonNull private final TemporalAmount minimumPeriod;
    @NonNull private final TemporalAmount maximumPeriod;
    @NonNull private final TemporalAmount suspensionPeriod;
    @NonNull private final boolean keepTimeOnMinimumPeriod;
    @NonNull private final boolean keepTimeOnMaximumPeriod;
    @NonNull private final boolean keepTimeOnSuspensionPeriod;

    public VaccinationPassportEvaluator(
        @NonNull final TemporalAmount minimumPeriod, final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriod, final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount suspensionPeriod, final boolean keepTimeOnSuspensionPeriod) {
        this.minimumPeriod = minimumPeriod;
        this.maximumPeriod = maximumPeriod;
        this.suspensionPeriod = suspensionPeriod;
        this.keepTimeOnMinimumPeriod = keepTimeOnMinimumPeriod;
        this.keepTimeOnMaximumPeriod = keepTimeOnMaximumPeriod;
        this.keepTimeOnSuspensionPeriod = keepTimeOnSuspensionPeriod;
    }

    /**
     * Evaluates vaccinations and lab results in the context of vaccination passport rules.
     * @return the result of the evaluation.
     */
    @NonNull
    @Override
    @SuppressWarnings({"java:S2637", "ConstantConditions"})
    public final PassportEvaluation evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results, @NonNull final LocalDateTime now) {
        setNow(now);
        final LocalDateTime completionDateTime = vaccination.completed();

        return firstNonNullOrDefaultSupplier(
            () -> grant(vaccinationValidUntil(completionDateTime)),
            // Reasons not to grant a vaccination passport follow below,
            // if none are met return above
            () -> denyForNotCompletedVaccination(completionDateTime),
            () -> denyForMinimumDurationAfterVaccinationHasNotPassed(completionDateTime),
            () -> denyForMaximumDurationAfterVaccinationHasPassed(completionDateTime),
            () -> denyForSuspendedDueToPositiveLabResult(results));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive lab results.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForSuspendedDueToPositiveLabResult(@NonNull final LabResults results) {
        return Objects.firstNonNullSupplier(
            () -> denyForPositivePcrResultWithinSuspensionPeriod(results),
            () -> denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(results));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive antigen result within the suspension period
     * that has not be set aside by a subsequent negative pcr result.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(@NonNull final LabResults results) {
        final LabResult.PositiveAntigenResult mostRecentPositiveAntigenResult = results.mostRecentPositiveAntigenResult();
        final LabResult.NegativePcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();

        if(mostRecentPositiveAntigenResult == null) return null;
        if(resultOutsideSuspensionPeriod(mostRecentPositiveAntigenResult)) return null;
        if(pcrResultAfterAntigenResult(mostRecentNegativePcrResult, mostRecentPositiveAntigenResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositiveAntigenResult), POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT);
    }

    /**
     * Evaluates if a passport should be denied because a vaccination is not completed.
     * @param completed the completion date of the vaccination, may be null if the vaccination was not completed.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private static PassportDeniedEvaluation denyForNotCompletedVaccination(@Nullable final LocalDateTime completed) {
        return completed == null ? denyIndefinitely(NOT_COMPLETED_VACCINATION) : null;
    }

    /**
     * Evaluates if a passport should be denied because the minimum duration after a vaccination has been completed has
     * not passed.
     * @param completed the completion date of the vaccination.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForMinimumDurationAfterVaccinationHasNotPassed(@NonNull final LocalDateTime completed) {
        final LocalDateTime vaccinationValidFrom = vaccinationValidFrom(completed);
        return now().isBefore(vaccinationValidFrom) ? deny(vaccinationValidFrom, MINIMUM_DURATION_NOT_PASSED): null;
    }

    /**
     * Evaluates if a passport should be denied because the maximum duration after a vaccination has been completed has
     * passed.
     * @param completed the completion date of the vaccination.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForMaximumDurationAfterVaccinationHasPassed(@NonNull final LocalDateTime completed) {
        final LocalDateTime validUntil = vaccinationValidUntil(completed);
        return now().isBefore(validUntil) ? null : denyIndefinitely(VACCINATION_EXPIRED);
    }

    /**
     * Evaluates if a passport should be denied because there is a positve pcr result within the suspension period.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation denyForPositivePcrResultWithinSuspensionPeriod(@NonNull final LabResults results) {
        final LabResult.PositivePcrResult mostRecentPositivePcrResult = results.mostRecentPositivePcrResult();

        if (mostRecentPositivePcrResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositivePcrResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositivePcrResult), POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD);
    }

    /**
     * Evaluates whether or not there a result is within the suspension period.
     * @param result the lab result to evaluate.
     * @return <code>false</code> if the result is within the suspension period, else <code>true</code>.
     */
    private boolean resultOutsideSuspensionPeriod(@NonNull final LabResult result) {
        final LocalDateTime resultTime = result.resultDateTime();
        return !resultTime.isAfter(startOfSuspensionPeriod());
    }

    /**
     * Evaluates whether or not an antigen result is before a pcr result.
     * @param pcrResult the most recent pcr result, can be <code>null</code>.
     * @param antigenResult the most recent positive antigen result.
     * @return <code>true</code> if the pcr result is not <code>null</code> and after the antigen result.
     */
    private static boolean pcrResultAfterAntigenResult(
        @Nullable final PcrResult pcrResult,
        @NonNull final AntigenResult antigenResult) {
        if(pcrResult == null) return false;

        final LocalDateTime dateTimeOfMostRecentPositiveAntigenResult = antigenResult.resultDateTime();
        final LocalDateTime dateTimeOfMostRecentNegativePcrResult = pcrResult.resultDateTime();
        return dateTimeOfMostRecentNegativePcrResult.isAfter(dateTimeOfMostRecentPositiveAntigenResult);
    }

    /**
     * Calculates the earliest date from which a completed vaccination can grant a passport
     * @param completed the completion date of the vaccination
     * @return the date
     */
    @NonNull
    private LocalDateTime vaccinationValidFrom(@NonNull LocalDateTime completed) {
        return resolutionStartOfDay(completed.plus(minimumPeriod), keepTimeOnMinimumPeriod);
    }

    /**
     * Calculates the latest date from which a completed vaccination can grant a passport
     * @param completed the completion date of the vaccination
     * @return the date
     */
    @NonNull
    private LocalDateTime vaccinationValidUntil(@NonNull final LocalDateTime completed) {
        return resolutionEndOfDay(completed.plus(maximumPeriod), keepTimeOnMaximumPeriod);
    }

    /**
     * Calculates the earliest day of a possible suspension of an completed vaccination due to positive results.
     * @return the date
     */
    @NonNull
    private LocalDateTime startOfSuspensionPeriod() {
        return resolutionStartOfDay(now().minus(suspensionPeriod), keepTimeOnSuspensionPeriod);
    }

    /**
     * Calculates the last day of a suspension of an completed vaccination due to positive results.
     * @return the date
     */
    @NonNull
    private LocalDateTime positiveResultExpiresSuspensionPeriod(@NonNull final LabResult.PositiveResult positiveResult) {
        return resolutionEndOfDay(positiveResult.resultDateTime().plus(suspensionPeriod), keepTimeOnSuspensionPeriod);
    }
}
