/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Results;
import dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations;
import org.junit.jupiter.params.provider.Arguments;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.Period;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAmount;

import static java.lang.String.format;

public final class ArgumentBuilder {
    public static final class Vaccine {
        public static final Period DEFAULT_MINIMUM_AMOUNT = Period.ofWeeks(2);
        public static final Period DEFAULT_MAXIMUM_AMOUNT = Period.ofMonths(6);
        public static final Period DEFAULT_SUSPENSION_AMOUNT = Period.ofDays(14);
    }
    public static final LocalDateTime DEFAULT_NOW = LocalDateTime.of(2000,1,15, 0,0, 0);
    private static final DateTimeFormatter DATE_FORMAT = DateTimeFormatter.ISO_DATE_TIME;

    @NonNull private final String name;
    private TemporalAmount minimumAmount;
    private boolean keepTimeOnMinimumAmount;
    private TemporalAmount maximumAmount;
    private boolean keepTimeOnMaximumAmount;
    private TemporalAmount suspensionAmount;
    private boolean keepTimeOnSuspensionAmount;
    private LocalDateTime now = DEFAULT_NOW;
    private Vaccination vaccination = Vaccinations.EMPTY;
    private LabResults labResults = Results.EMPTY;
    private boolean granted;
    private LocalDateTime expire;

    private ArgumentBuilder(@NonNull final String name) {
        this.name = name;
    }

    @NonNull
    public static ArgumentBuilder arguments(@NonNull final String name) {
        return new ArgumentBuilder(name);
    }

    @NonNull
    public static ArgumentBuilder defaultForVaccinePass(@NonNull final String name) {
        return arguments(name)
            .minimumAmount(Vaccine.DEFAULT_MINIMUM_AMOUNT)
            .maximumAmount(Vaccine.DEFAULT_MAXIMUM_AMOUNT)
            .suspensionAmount(Vaccine.DEFAULT_SUSPENSION_AMOUNT)
            .keepTimeOnMinimumAmount()
            .keepTimeOnMaximumAmount()
            .keepTimeOnSuspensionAmount();
    }

    @NonNull
    public ArgumentBuilder minimumAmount(@NonNull final TemporalAmount minimumAmount) {
        this.minimumAmount = minimumAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder maximumAmount(@NonNull final TemporalAmount maximumAmount) {
        this.maximumAmount = maximumAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder suspensionAmount(@NonNull final TemporalAmount suspensionAmount) {
        this.suspensionAmount = suspensionAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder keepTimeOnMinimumAmount(final boolean keepTimeOnMinimumAmount) {
        this.keepTimeOnMinimumAmount = keepTimeOnMinimumAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder keepTimeOnMaximumAmount(final boolean keepTimeOnMaximumAmount) {
        this.keepTimeOnMaximumAmount = keepTimeOnMaximumAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder keepTimeOnSuspensionAmount(final boolean keepTimeOnSuspensionAmount) {
        this.keepTimeOnSuspensionAmount = keepTimeOnSuspensionAmount;
        return this;
    }

    @NonNull
    public ArgumentBuilder keepTimeOnMinimumAmount() {
        return keepTimeOnMinimumAmount(true);
    }

    @NonNull
    public ArgumentBuilder keepTimeOnMaximumAmount() {
        return keepTimeOnMaximumAmount(true);
    }

    @NonNull
    public ArgumentBuilder keepTimeOnSuspensionAmount() {
        return keepTimeOnSuspensionAmount(true);
    }

    @NonNull
    public ArgumentBuilder vaccination(@NonNull final Vaccination vaccination) {
        this.vaccination = vaccination;
        return this;
    }

    @NonNull
    public ArgumentBuilder labResults(@NonNull final LabResults labResults) {
        this.labResults = labResults;
        return this;
    }

    @NonNull
    public ArgumentBuilder now(@NonNull final LocalDateTime now) {
        this.now = now;
        return this;
    }

    @NonNull
    private ArgumentBuilder granted(final boolean granted) {
        this.granted = granted;
        return this;
    }

    @NonNull
    public ArgumentBuilder granted() {
        return granted(true);
    }

    @NonNull
    public ArgumentBuilder denied() {
        return granted(false);
    }

    @NonNull
    public ArgumentBuilder expire(@NonNull final LocalDateTime expire) {
        this.expire = expire;
        return this;
    }

    @NonNull
    public ArgumentBuilder expireNever() {
        this.expire = null;
        return this;
    }

    @NonNull
    public Arguments build() {
        return Arguments.arguments(
            granted ? "GRANT" : "DENY",
            name,
            friendly(),
            minimumAmount,
            keepTimeOnMinimumAmount,
            maximumAmount,
            keepTimeOnMaximumAmount,
            suspensionAmount,
            keepTimeOnSuspensionAmount,
            vaccination,
            labResults,
            now,
            granted,
            expire);
    }

    @NonNull
    private String friendly() {
        return format(
            "{ min = %s, keep = %s, max = %s, keep = %s, susp = %s, keep = %s, vac = %s, lab = %s, now = %s, granted = %s, expire = %s }",
            minimumAmount,
            keepTimeOnMinimumAmount,
            maximumAmount,
            keepTimeOnMaximumAmount,
            suspensionAmount,
            keepTimeOnSuspensionAmount,
            friendly(vaccination),
            friendly(labResults),
            DATE_FORMAT.format(now),
            granted,
            expire == null ? null : DATE_FORMAT.format(expire));
    }

    private String friendly(@NonNull final LabResults labResults) {
        return "{ " +
            labResults.results().stream().reduce(
                "",
                (a, l) -> "".equals(a) ? friendly(l) : "," + friendly(l),
                (x, y) -> x + "," + y
            ) +
            " }";
    }

    private String friendly(@NonNull final LabResult labResult) {
        return format(
            "{ time = %s, type = %s, result = %s }",
            DATE_FORMAT.format(labResult.resultDateTime()),
            labResult instanceof LabResult.PcrResult ? "pcr" : "ag",
            labResult.positive() ? "pos" : "neg");
    }

    private String friendly(@NonNull final Vaccination vaccination) {
        final LocalDateTime completed = vaccination.completed();
        return format(
            "{ completed = %s }",
            completed == null ? null : DATE_FORMAT.format(completed));
    }
}
