/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.time.temporal.TemporalAmount;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.DEFAULT_NOW;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MAXIMUM_AMOUNT;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_MINIMUM_AMOUNT;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.Vaccine.DEFAULT_SUSPENSION_AMOUNT;
import static dk.sundhedsdatastyrelsen.cbs.rules.ArgumentBuilder.defaultForVaccinePass;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positiveAntigen;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positiveAntigenNegativePcr;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Results.positivePcr;
import static dk.sundhedsdatastyrelsen.cbs.test.Data.Vaccinations.completed;
import static org.junit.jupiter.api.Assertions.assertEquals;

class VaccinationPassportEvaluatorTest {

    @SuppressWarnings("unused")
    @DisplayName("evaluation")
    @ParameterizedTest(name = "[{0}] - {1} : {2}")
    @MethodSource("evaluateSource")
    void evaluate(
        @NonNull final String outcome,
        @NonNull final String name,
        @NonNull final String friendly,
        @NonNull final TemporalAmount minimumPeriod,
        final boolean keepTimeOnMinimumPeriod,
        @NonNull final TemporalAmount maximumPeriod,
        final boolean keepTimeOnMaximumPeriod,
        @NonNull final TemporalAmount suspensionPeriod,
        final boolean keepTimeOnSuspensionPeriod,
        @NonNull final Vaccination vaccination,
        @NonNull final LabResults labResults,
        @NonNull final LocalDateTime now,
        @NonNull final boolean granted,
        @NonNull final LocalDateTime validUntil) {
        final PassportEvaluator evaluator = new VaccinationPassportEvaluator(minimumPeriod, keepTimeOnMinimumPeriod, maximumPeriod, keepTimeOnMaximumPeriod, suspensionPeriod, keepTimeOnSuspensionPeriod);
        final PassportEvaluation evaluation = evaluator.evaluate(vaccination, labResults, now);
        assertEquals(granted, evaluation.passportGranted());
        assertEquals(validUntil, evaluation.expiresAfter());
    }

    @NonNull
    private static Stream<Arguments> evaluateSource() {
        return Stream.of(
            defaultForVaccinePass("No vaccinations").denied().build(),
            defaultForVaccinePass("Completed vaccination where less than minimum period has passed").vaccination(completed(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MINIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where less than maximum period has passed").vaccination(completed(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT, true))).granted().expire(expireGranted(beforeMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly maximum period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT, true))).denied().expireNever().build(),
            defaultForVaccinePass("Completed vaccination where more than maximum period has passed").vaccination(completed(afterMax(DEFAULT_NOW, DEFAULT_MAXIMUM_AMOUNT, true))).denied().expireNever().build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive pcr where more than suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positivePcr(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive pcr where exactly suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positivePcr(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive pcr where less suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positivePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true), DEFAULT_SUSPENSION_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive antigen where more than suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positiveAntigen(beforeMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive antigen where exactly suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positiveAntigen(on(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positiveAntigen(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).denied().expire(expireDenied(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true), DEFAULT_SUSPENSION_AMOUNT, true)).build(),
            defaultForVaccinePass("Completed vaccination where exactly minimum period has passed and positive antigen where less suspension period has passed and negative pcr test after positive antigen").vaccination(completed(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true))).labResults(positiveAntigenNegativePcr(afterMin(DEFAULT_NOW, DEFAULT_SUSPENSION_AMOUNT, true))).granted().expire(expireGranted(on(DEFAULT_NOW, DEFAULT_MINIMUM_AMOUNT, true), DEFAULT_MAXIMUM_AMOUNT, true)).build());
    }

    @NonNull
    private static LocalDateTime beforeMin(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.minus(amount).minusNanos(1) : time.minus(amount).minusDays(1);
    }

    @NonNull
    private static LocalDateTime beforeMax(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.minus(amount).plusNanos(1) : time.minus(amount).plusDays(1);
    }

    @NonNull
    private static LocalDateTime afterMin(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.minus(amount).plusNanos(1) : time.minus(amount).plusDays(1);
    }

    @NonNull
    private static LocalDateTime afterMax(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.minus(amount).minusNanos(1) : time.minus(amount).minusDays(1);
    }

    @NonNull
    private static LocalDateTime on(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.minus(amount) : time.minus(amount);
    }

    @NonNull
    private static LocalDateTime expireDenied(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.plus(amount) : time.plus(amount).plusDays(1).toLocalDate().atStartOfDay();
    }

    @NonNull
    private static LocalDateTime expireGranted(@NonNull final LocalDateTime time, @NonNull final TemporalAmount amount, final boolean keepTime) {
        return keepTime ? time.plus(amount) : time.plus(amount).plusDays(1).toLocalDate().atStartOfDay();
    }

}
