/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.config;

import dk.sundhedsdatastyrelsen.cbs.service.security.SignerFactory;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Profile;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.io.Resource;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import se.digg.dgc.service.DGCEncoder;
import se.digg.dgc.service.impl.DefaultDGCEncoder;
import se.digg.dgc.service.impl.DefaultDKEncoder;
import se.digg.dgc.signatures.DGCSigner;
import se.digg.dgc.signatures.impl.DefaultDKSigner;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.Security;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;

@Configuration
@PropertySource("classpath:application-service.properties")
public class SignatureConfiguration {

    private static final Logger LOGGER = Logger.getLogger(SignatureConfiguration.class);

    private static final String LUNA_PROVIDER_NAME = "LunaProvider";
    private static final String LUNA_KEY_STORE_TYPE = "Luna";

    @Bean
    public DGCEncoder dgcEncoder(DGCSigner signer) {
        DefaultDGCEncoder defaultDGCEncoder = new DefaultDGCEncoder(signer);
        defaultDGCEncoder.setTransliterateNames(true);
        return defaultDGCEncoder;
    }

    @Bean
    public DefaultDKEncoder dkEncoder(DefaultDKSigner signer) {
        return new DefaultDKEncoder(signer);
    }

    @Bean
    public DGCSigner dgcSigner(SignerFactory factory) throws CertificateException {
        return factory.dgcSigner();
    }

    @Bean
    public DefaultDKSigner defaultDKSigner(SignerFactory factory) throws CertificateException {
        return factory.dkSigner();
    }

    @Bean
    public SignerFactory signerFactory(
        @NonNull KeyStore keyStore,
        @Nullable Provider provider,
        @Value("${signing.key-store.key.alias:covid-19 borgerstatus (funktionscertifikat)}") String alias,
        @Value("${signing.key-store.key.password:Test1234}") char[] password
    ) throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException {
        return new SignerFactory(keyStore, alias, password, provider);
    }

    @Bean
    @Profile("local-key-store")
    public KeyStore localKeyStore(
        @Value("${signing.key-store.path:Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12}") Resource keyStoreResource,
        @Value("${signing.key-store.password:Test1234}") String password
    ) throws KeyStoreException, CertificateException, IOException, NoSuchAlgorithmException {
        KeyStore keystore = KeyStore.getInstance("PKCS12");
        try (InputStream inputStream = keyStoreResource.getInputStream()) {
            keystore.load(inputStream, password.toCharArray());
        }
        return keystore;
    }

    @Bean
    @Profile("!local-key-store")
    public KeyStore lunaKeyStore(
        @Value("${signing.key-store.slot:0}") int slot,
        @Value("${signing.key-store.password:Test1234}") String password
    ) throws KeyStoreException, CertificateException, IOException, NoSuchAlgorithmException {
        LOGGER.debug("Loading Luna KeyStore");
        KeyStore keystore = KeyStore.getInstance(LUNA_KEY_STORE_TYPE);
        try (ByteArrayInputStream inputStream = new ByteArrayInputStream(String.format("slot:%d%nusertype:CKU_CRYPTO_USER%n", slot).getBytes(StandardCharsets.UTF_8))) {
            keystore.load(inputStream, password.toCharArray());
            LOGGER.debug("Luna KeyStore loaded");
        }
        return keystore;
    }

    @Bean
    @Profile("!local-key-store")
    public Provider lunaProvider() {
        Provider provider = Security.getProvider(LUNA_PROVIDER_NAME);
        if (provider == null) {
            throw new IllegalStateException("Luna Provider is not present");
        }
        return provider;
    }

}
