/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Strings;
import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.VaccinationPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvaluation;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.KeyValuePair;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassList;
import org.springframework.stereotype.Service;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.payload.v1.PersonName;
import se.digg.dgc.service.DGCEncoder;
import se.digg.dgc.service.impl.DefaultDKEncoder;

import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static no.kantega.id.dk.DanishIdNumber.birthday;
import static no.kantega.id.dk.DanishIdNumber.forId;

@Service
public class PassportFactory {


    private final DGCEncoder euEncoder;
    private final DefaultDKEncoder dkEncoder;
    private final VaccineMapper vaccineMapper;
    private final VaccinationPassportEvaluator vaccinationValidator;
    private final Duration maximumDuration;

    public PassportFactory(DGCEncoder euEncoder, DefaultDKEncoder dkEncoder, VaccineMapper vaccineMapper, VaccinationPassportEvaluator vaccinationValidator) {
        this.euEncoder = euEncoder;
        this.dkEncoder = dkEncoder;
        this.vaccineMapper = vaccineMapper;
        this.vaccinationValidator = vaccinationValidator;
        this.maximumDuration = Duration.ofDays(3);
    }

    private PassJWT getDKPlusPass(PersonRecord person, Instant expiry) {

        PassJWT passJWT = new PassJWT();
        passJWT.setType("DKPLUS");

        LocalDate date = cprToDate(person.personIdentifier());
        String name = getFirstname(person) + " " +  person.surName();

        DKPlus dkPlus = new DKPlus();
        dkPlus.setLegalName(name);
        dkPlus.setDateOfBirth(DateTimeFormatter.ISO_DATE.format(date));

        try {
            String json = new ObjectMapper().writeValueAsString(dkPlus);
            String encoded = dkEncoder.encodeDkPlus(CBORObject.FromJSONString(json), expiry, UUID.randomUUID().toString());
            passJWT.setJWT(encoded);
            return passJWT;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    private PassJWT getDKPass(Instant expiry) {
        PassJWT passJWT = new PassJWT();
        passJWT.setType("DK");
        try {
            passJWT.setJWT(dkEncoder.encodeDk(CBORObject.NewMap(), expiry, UUID.randomUUID().toString()));
            return passJWT;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    private List<PassJWT> buildDKpassports(PersonRecord person, Instant expiry) {
        return Arrays.asList(getDKPass(expiry), getDKPlusPass(person, expiry));
    }

    public GetCoronaPassResponse createPassportFrom(Vaccination vaccination, LabResults labResults, PersonRecord person) {

        GetCoronaPassResponse coronaPassResponse = new GetCoronaPassResponse();
        coronaPassResponse.setCoronaPasses(new PassList());

        PassportEvaluation result = vaccinationValidator.evaluate(vaccination, labResults, LocalDateTime.now());

        if (result.passportGranted()) {
            Instant expiry = getExpiry(result);
            coronaPassResponse.getCoronaPasses().getPass().addAll(buildDKpassports(person, expiry));
            coronaPassResponse.getCoronaPasses().getPass().addAll(buildEUpassports(person, vaccination, labResults, expiry));
        }
        else
        {
            //TODO fix expiry to response from https://jira.nspop.dk/browse/CR-117
            coronaPassResponse.getCoronaPasses().getPass().addAll(buildEUpassports(person, vaccination, labResults, Instant.now().plus(maximumDuration)));
        }
        return coronaPassResponse;
    }

    private Instant getExpiry(PassportEvaluation result) {
        //TODO use 'Instant' once available from the result
        Instant expiry = Instant.now().plus(maximumDuration);
        if (expiry.isAfter(result.expiresAfter().toLocalDate().atStartOfDay(ZoneId.systemDefault()).toInstant()))
            expiry = result.expiresAfter().toLocalDate().atStartOfDay(ZoneId.systemDefault()).toInstant();
        return expiry;
    }

    private Collection<? extends PassJWT> buildEUpassports(PersonRecord person, Vaccination vaccination, LabResults labResults, Instant defaultExpiry) {

        return Stream.of(
            labResults.mostRecentResult() != null ? addLabResultsAndEncode(buildBaselineCertificateFrom(person), labResults, defaultExpiry) : null,
            vaccination.getLatestEffectuated() != null ? addVaccineAndEncode(buildBaselineCertificateFrom(person), vaccination, defaultExpiry) : null
        )
            .filter(Objects::nonNull).collect(Collectors.toList());
    }

    private PassJWT addVaccineAndEncode(DigitalGreenCertificate certificate, Vaccination vaccination, Instant defaultExpiry) {
        certificate.setV(Collections.singletonList(vaccineMapper.toVac(vaccination)));
        return encode(certificate, "vac", defaultExpiry);
    }


    private PassJWT addLabResultsAndEncode(DigitalGreenCertificate certificate, LabResults labResults, Instant defaultExpiry) {
        certificate.setT(Collections.singletonList(LabResultMapper.toTestEntry(labResults.mostRecentResult())));
        return encode(certificate, "lab", defaultExpiry);
    }


    private PassJWT encode(DigitalGreenCertificate certificate, String value, Instant defaultExpiry) {
        PassJWT passJWT = new PassJWT();
        passJWT.setType("EU");
        try {
            passJWT.setJWT(euEncoder.encode(certificate, defaultExpiry));
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
        KeyValuePair kv = new KeyValuePair();
        kv.setKey("eu:keytype");
        kv.setValue(value);
        passJWT.getProperties().add(kv);
        return passJWT;
    }

    private static DigitalGreenCertificate buildBaselineCertificateFrom(PersonRecord person) {
        DigitalGreenCertificate certificate = new DigitalGreenCertificate();
        certificate.setNam(new PersonName()
            .withFn(getFirstname(person))
            .withGn(person.surName()));
        certificate.setDob(cprToDate(person.personIdentifier()));
        certificate.setVer("1.0.0");
        return certificate;
    }

    private static String getFirstname(PersonRecord person) {
        return Stream.of(person.firstName(), person.middleName())
            .filter(Objects::nonNull)
            .filter(s -> !Strings.isNullOrEmpty(s))
            .collect(Collectors.joining(" "));
    }

    private static LocalDate cprToDate(String cpr) {
        return birthday(forId(cpr)).get();
    }
}
