/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import com.google.common.collect.MoreCollectors;
import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext.Ticket;
import dk.sds.nsp.security.SecurityProvider;
import dk.sundhedsdatastyrelsen.cbs.dao.PersonRecord;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.lab.LabResultsBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.cbs.test.ServiceConfiguration;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.lang.NonNull;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.impl.DefaultDGCDecoder;
import se.digg.dgc.service.impl.MrzParser;
import se.digg.dgc.signatures.CertificateProvider;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;

import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.UUID;
import java.util.stream.Collectors;

import static dk.sds.nsp.security.SecurityContext.User;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@CbsTest
@SpringJUnitConfig({ServiceConfiguration.class})
class CoronaPassWebServiceTest {

    private static final long DRUG_ID = 32138956351L;

    enum Content {
        DK,
        DKPLUS
    }

    private static Certificate certificate;
    private static DefaultDGCSignatureVerifier verifier;
    private static CertificateProvider certificateProvider;

    private final ApplicationContext applicationContext = new AnnotationConfigApplicationContext("dk.sundhedsdatastyrelsen.cbs.test");
    private final CoronaPassWebService coronaPassWebService = new CoronaPassWebService(applicationContext);
    private final RandomData randomData = new RandomData(new Random().nextInt());

    @NonNull
    private final Reader<Vaccination, String> vaccinationReader;
    @NonNull
    private final Reader<LabResults, String> labResultReader;
    @NonNull
    private final Reader<PersonRecord, String> personRecordReader;
    @NonNull
    private final User user;

    @Autowired
    CoronaPassWebServiceTest(
        @NonNull final Reader<Vaccination, String> vaccinationReader,
        @NonNull final Reader<LabResults, String> labResultReader,
        @NonNull final Reader<PersonRecord, String> personRecordReader) {
        this.vaccinationReader = vaccinationReader;
        this.labResultReader = labResultReader;
        this.personRecordReader = personRecordReader;
        user = mock(User.class);
        final DevelopmentSecurityProvider securityProvider = (DevelopmentSecurityProvider) SecurityProvider.provider();
        securityProvider.setGlobalTicket(when(mock(Ticket.class).isValid()).thenReturn(true).getMock());
        securityProvider.setGlobalActingUser(user);

    }

    @BeforeAll
    public static void beforeAll() throws KeyStoreException, CertificateException, IOException, NoSuchAlgorithmException {
        KeyStore keystore = KeyStore.getInstance("PKCS12");
        keystore.load(CoronaPassWebServiceTest.class.getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());
        certificate = keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
        verifier = new DefaultDGCSignatureVerifier();
        certificateProvider = (country, kid) -> {
            List<X509Certificate> listOfCertificate = new ArrayList<>();
            listOfCertificate.add((X509Certificate) certificate);
            return listOfCertificate;
        };
    }

    @NonNull
    private GetCoronaPassRequest request(String cpr) {
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(cpr);
        when(user.getIdentifier()).thenReturn(cpr);
        return request;
    }

    @NonNull

    private MockDataService.Person withMockedPerson(@NonNull final String cpr) {

        MockDataService.Person person = MockDataService.person(cpr);
        PersonRecord record = mock(PersonRecord.class);
        when(record.firstName()).thenReturn(person.first);
        when(record.middleName()).thenReturn("");
        when(record.surName()).thenReturn(person.last);
        when(record.personIdentifier()).thenReturn(cpr);
        when(personRecordReader.search(cpr)).thenReturn(record);
        return person;
    }


    @NonNull
    private String cprWithCacheHitOn(@NonNull final String cpr) {

        MockDataService.Person person = MockDataService.person(cpr);
        PersonRecord record = mock(PersonRecord.class);
        when(record.firstName()).thenReturn(person.first);
        when(record.middleName()).thenReturn("");
        when(record.surName()).thenReturn(person.last);
        when(record.personIdentifier()).thenReturn(cpr);
        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.vaccination(cpr, true).add(
            effectuatedVaccination(32138956351L,DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), LocalDateTime.now().minus(336, ChronoUnit.HOURS))
                .build()).build());
        when(labResultReader.search(cpr)).thenReturn(mock(LabResults.class));
        when(personRecordReader.search(cpr)).thenReturn(record);
        return cpr;
    }

    @NonNull
    private String cprWithCacheMissOn(@NonNull final String cpr) {
        when(vaccinationReader.search(cpr)).thenReturn(null);
        when(labResultReader.search(cpr)).thenReturn(null);
        return cpr;
    }

    @Test
    void test_getCoronaPass_cache_hit() {
        final String cprPrefix = "040587010";
        for (int i = 0; i < 10; i++) {
            final String cpr = cprWithCacheHitOn(cprPrefix + 1);
            verifyPass(cpr);
        }
    }

    @Test
    void test_get_no_passes_case_0405870114() {
        String cpr = "0405870114";
        PersonRecord record = mock(PersonRecord.class);

        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.vaccination(cpr, false).build());
        when(labResultReader.search(cpr)).thenReturn(mock(LabResults.class));
        when(personRecordReader.search(cpr)).thenReturn(record);

        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNull(coronaPass.getFutureResult());
        assertEquals(0, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    @Disabled("Until cdj adds the TestValidator")
    void test_verify_DK_and_verify_DKPlus() {

        String cpr = "0405870113";
        MockDataService.Person person = withMockedPerson(cpr);
        withMockedVaccine(cpr);
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNull(coronaPass.getFutureResult());

        Cwt dkContent = getCwt(coronaPass, Content.DK);
        Cwt dkPlusContent = getCwt(coronaPass, Content.DKPLUS);

        assertTrue(timestampIsApproxNow(dkContent.getIssuedAt()));
        assertTrue(timestampIsApprox72HoursAhead(dkContent.getExpiration()));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(timestampIsApproxNow(dkPlusContent.getIssuedAt()));
        assertTrue(timestampIsApprox72HoursAhead(dkPlusContent.getExpiration()));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertEquals("1987-05-04", dkPlusContent.getClaim("dateOfBirth").AsString());
        assertEquals(person.first + " " + person.last, dkPlusContent.getClaim("legalName").AsString());
    }

    private void withMockedLabResult(String cpr, LabResults labResults) {
        when(labResultReader.search(cpr)).thenReturn(labResults);
    }

    private void withMockedVaccine(String cpr) {
        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.vaccination(cpr, true).add(
            effectuatedVaccination(1L, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), LocalDateTime.now().minus(336, ChronoUnit.HOURS))
                .build()).build());
    }

    private void withMockedVaccine(String cpr, EffectuatedVaccination effectuatedVaccination) {
        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.vaccination(cpr, true).add(effectuatedVaccination).build());
    }

    private void withNoVaccine(String cpr) {
        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.vaccination(cpr, false).build());
    }

    private void withNoLabResults(String cpr) {
        when(labResultReader.search(cpr)).thenReturn(LabResultsBuilder.cpr(cpr).build());
    }


    @Test
    void test_getCoronaPassOnlyEU() {
        String cpr = cprWithCacheHitOn("0405870115");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPass);
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getNam().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getNam().getFn(), -1);
            assertEquals(expectedGnt, certificate.getNam().getGnt());
            assertEquals(expectedFnt, certificate.getNam().getFnt());
        });
    }

    @Test
    void test_getCoronaPassEUMRZnames() {
        String cpr = cprWithCacheHitOn("2011451636");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPass);
        assertEquals(1, listDGC.size());
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getNam().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getNam().getFn(), -1);
            assertEquals(expectedGnt, certificate.getNam().getGnt());
            assertEquals(expectedFnt, certificate.getNam().getFnt());
        });
    }

    @Test
    @Disabled("Until cdj adds the TestValidator")
    void test_getCoronaPass5minutesExpiry() {
        String cpr = cprWithCacheHitOn("0405870118");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getPayload(eu).getExpiration();
        assertTrue(Instant.now().plus(4, ChronoUnit.MINUTES).isBefore(expDate));
        assertTrue(Instant.now().plus(6, ChronoUnit.MINUTES).isAfter(expDate));
    }

    @Test
    void test_verify_signature() throws Exception {
        String cpr = cprWithCacheHitOn("0405870118");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        DigitalGreenCertificate certificate = getEUPayload(eu.getJWT());
        assertFalse(eu.getProperties().isEmpty());
    }

    @Test
    @Disabled("Until cdj adds the TestValidator")
    void test_getCoronaPass_15_minutes_to_expiry_based_on_PCR_LabResults() {
        String cpr = "0405870117";

        withMockedPerson("0405870117");
        withNoVaccine(cpr);
        withLabResultIssuedAt(cpr, LocalDateTime.now().minus((24 * 3 * 60) - 15, ChronoUnit.MINUTES));

        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNull(coronaPass.getFutureResult());

        verifyDKContent(coronaPass);

        List<Cwt> rawCwts = payloadToRawEUCWT(coronaPass);
        assertEquals(1, rawCwts.size());

        rawCwts.stream().forEach(c -> {
            assertTrue(timestampIsApproxNow(c.getExpiration()));
        });
    }

    @Test
    @Disabled("Used for manual inspection")
    void parseToken()
    {
        String pass = "HC1:NCFZ60/F0A2WF.9.60+O4W79F%GZM9NTVO KFBBPG3*70NS8FN0I9CNNJWY0E9C./KD97TK0H90GECFAGZWE-CLOPCU8F6%E3.D9%EPED5G7D%6/96B46+TCSW65VC8W5TZC346Y57NF6BX5-96D%6A56KA72UCVX6UPC1JC%N9JVC1VE846Y96A465W50M6$965G7G466468JB+EDD8F0T85/DOQET34X C:VDG7D0T83WEM348%E59D*KERWE3KCZPCNF6OF63W56L6%96.SAPA7.961G73564KCD3DX47B46IL6646H*6KWEKDDC%6-Q6QW66464KCAWE6T9G%6G%67W5KECHAGGVC*JC1A67463W5Y96RF6TPCBEC7ZKW.C$OC..D JC6/D9O9+ED7WEF/D1$CLPCG/D$OC8+9MPCG/DVJD5IA2OACY8WNA.+9O/EZKEZ96446256V50BU9B4S:/E0FST DIZ8-WCU4GE+5Z3AXJSE$T5L8*VKB5G/R3W1FC/I*NS$/R9/O0Q158SICHSYPNOAF4ICWH8YJJTEG$L9VTP:H2VQ5VG3HQ GHXW5%V1:FIPXLGXJRKKMT882ICNHWWSX/BK5NA8QX6FOLKSKR%94:9D*7FIYH5V03EDXTKFY2%6EH52+V48*8%JLW2OY4HB.UUI0*EVUJDYOAMLHP85P0D26C3/B%-EEO8BAMBSV$W6P4OY0FN7A6/V3ROM%V1YCXB4:29W:NYDUB1F*34QH9AYJ1E3VKM8IF6G1TVC%PBI88PH4AGC2$9EHMNDAC-47X9K*1ZD7Y-5EMK%PL87HYUJ$V4*BGHE6T/9DR9537FEAC$3SLA%K7- S";
        DigitalGreenCertificate cert = getEUPayload(pass.substring(4));
        System.out.println(cert);
    }

    private void withLabResultIssuedAt(String cpr, LocalDateTime resultDateTime) {
        LabResult.PcrResult pcrResult = mock(LabResult.PcrResult.class);
        when(pcrResult.negative()).thenReturn(false);
        when(pcrResult.resultDateTime()).thenReturn(resultDateTime);
        when(pcrResult.samplingDateTime()).thenReturn(resultDateTime.plus(15, ChronoUnit.MINUTES));
        when(pcrResult.id()).thenReturn(randomData.nextLong());
        when(pcrResult.type()).thenReturn("SARS2");
        when(pcrResult.center()).thenReturn(randomData.uuidString());
        withMockedLabResult(cpr, LabResultsBuilder.cpr(cpr).add(pcrResult).build());
    }

    @Test
    void test_getCoronaPass_72_hour_expiry_based_on_vaccination() {
        String cpr = "0405870107";

        withMockedPerson("0405870107");
        LocalDateTime effectuatedAt =LocalDateTime.ofInstant( Instant.now().minus(177 * 24, ChronoUnit.HOURS), ZoneOffset.UTC);
        withMockedVaccine(cpr, EffectuatedVaccinationBuilder.effectuatedVaccination(32139944147L, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuatedAt).build());
        withNoLabResults(cpr);

        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNull(coronaPass.getFutureResult());

        verifyDKContent(coronaPass);

        List<Cwt> rawCwts = payloadToRawEUCWT(coronaPass);
        assertEquals(1, rawCwts.size());

        rawCwts.stream().forEach(c -> {
            timestampIsApprox72HoursAhead(c.getExpiration());
        });
    }


    private void verifyPass(String cpr) {
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        final String cpr = cprWithCacheMissOn("1234567890");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNotNull(coronaPass.getFutureResult());
        assertNotNull(coronaPass.getFutureResult().getJobId());
        assertNull(coronaPass.getCoronaPasses());
    }


    private void verifyDKContent(GetCoronaPassResponse coronaPass) {
        Cwt dkContent = getCwt(coronaPass, Content.DK);
        Cwt dkPlusContent = getCwt(coronaPass, Content.DKPLUS);

        assertTrue(dkContent.getExpiration().isAfter(Instant.now()));
        assertTrue(dkContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(dkPlusContent.getExpiration().isAfter(Instant.now()));
        assertTrue(dkPlusContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertNotNull(dkPlusContent.getClaim("dateOfBirth"));
        assertNotNull(dkPlusContent.getClaim("legalName"));
    }

    private Cwt getCwt(GetCoronaPassResponse coronaPass, Content content) {
        if (content == Content.DK) {
            PassJWT dkPass = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "DK".equals(p.getType())).collect(MoreCollectors.onlyElement());
            assertTrue(dkPass.getJWT().startsWith("DK1:"));
            return getPayload(dkPass);
        } else if (content == Content.DKPLUS) {
            PassJWT dkPlusPass = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "DKPLUS".equals(p.getType())).collect(MoreCollectors.onlyElement());
            assertTrue(dkPlusPass.getJWT().startsWith("DK2:"));
            return getPayload(dkPlusPass);
        }
        throw new RuntimeException();
    }


    private Cwt getPayload(PassJWT eu) {
        return getPayload(eu.getJWT().substring(4));
    }

    private static Cwt getPayload(String token) {
        try {
            byte[] cborBytes = Zlib.decompress(Base45.getDecoder().decode(token), true);
            CoseSign1_Object coseObject = new CoseSign1_Object(cborBytes);
            coseObject.verifySignature(certificate.getPublicKey());
            return coseObject.getCwt();
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    public static List<DigitalGreenCertificate> payloadToEUpasses(GetCoronaPassResponse passes) {
        return passes.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).map(PassJWT::getJWT).map(s -> s.substring(4)).map(CoronaPassWebServiceTest::getEUPayload).collect(Collectors.toList());
    }

    public static List<Cwt> payloadToRawEUCWT(GetCoronaPassResponse passes) {
        return passes.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).map(PassJWT::getJWT).map(s -> s.substring(4)).map(CoronaPassWebServiceTest::getPayload).collect(Collectors.toList());
    }

    public static DigitalGreenCertificate getEUPayload(String token) {
        try {
            return new DefaultDGCDecoder(verifier, certificateProvider).decode(token);
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    boolean timestampInRange(Instant start, Instant end, Instant subject) {
        return !subject.isBefore(start) && !subject.isAfter(end);
    }

    boolean timestampIsApproxNow(Instant subject) {
        return timestampInRange(Instant.now().minus(1, ChronoUnit.SECONDS), Instant.now().plus(1, ChronoUnit.SECONDS), subject);
    }

    boolean timestampIsApprox72HoursAhead(Instant subject) {
        return timestampInRange(Instant.now().plus(71, ChronoUnit.HOURS).plus(59, ChronoUnit.MINUTES), Instant.now().plus(72, ChronoUnit.HOURS).plus(1, ChronoUnit.MINUTES), subject);
    }

}
