/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.EffectuatedVaccination;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import se.digg.dgc.payload.v1.VaccinationEntry;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.EffectuatedVaccinationBuilder.effectuatedVaccination;
import static dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder.vaccination;
import static java.time.LocalDateTime.now;
import static org.junit.jupiter.api.Assertions.*;

class VaccineMapperTest {

    private VaccineMapper vaccineMapper;
    private VaccineMappingConfiguration vaccineMappingConfiguration;
    private final static long DRUG_ID = 1;


    @BeforeEach
    void beforeEach() {
        Map<Long, VaccineMappingConfiguration.VaccineInformation> mappingConfiguration = new HashMap<>();
        mappingConfiguration.put(1L, new VaccineMappingConfiguration.VaccineInformation(UUID.randomUUID().toString(), UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        vaccineMappingConfiguration = new VaccineMappingConfiguration(mappingConfiguration);
        vaccineMapper = new VaccineMapper(vaccineMappingConfiguration);
    }

    @Test
    void testIncompleteConfiguration() {
        int vaccineId = 2;
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
            .build();
        VaccineMapperException exception = assertThrows(VaccineMapperException.class, () -> vaccineMapper.toVac(vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build()));
        assertTrue(exception.getMessage().contains("Unknown vaccineIdentifier: '2' - vaccine cannot be mapped"));
    }

    @Test
    void mapCompletedSingleVaccine() {

        int vaccineId = 1;
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true).add(effectuatedVaccination).build();

        VaccinationEntry actualVac = vaccineMapper.toVac(vaccination);
        assertEquals(vaccination.completed().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(1, actualVac.getDn());
        assertEquals(1, actualVac.getSd());
        assertEquals("Danish Health Data Authority", actualVac.getIs());
        assertNotNull(UUID.fromString(actualVac.getCi().replace("urn:uuid:", "")));
    }

    @Test
    void mapCompletedVaccineWithTwoEffectuations() {

        int vaccineId = 1;
        LocalDateTime effectuated = now();

        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), true)
            .add(effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
                .build())
            .add(effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
                .build()).build();

        VaccinationEntry actualVac = vaccineMapper.toVac(vaccination);
        assertEquals(vaccination.completed().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(2, actualVac.getDn());
        assertEquals(2, actualVac.getSd());

    }

    @Test
    void mapIncompletedVaccineWithOneEffectuatedVaccine() {

        int vaccineId = 1;
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false).add(effectuatedVaccination).build();

        VaccinationEntry actualVac = vaccineMapper.toVac(vaccination);
        assertEquals(vaccination.getLatestEffectuated().effectuated().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(1, actualVac.getDn());
        assertEquals(2, actualVac.getSd());
    }

    @Test
    void mapIncompletedVaccineWithTwoEffectuatedVaccine() {

        int vaccineId = 1;
        LocalDateTime effectuated = now();

        final EffectuatedVaccination effectuatedVaccination = effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated)
            .build();
        Vaccination vaccination = vaccination(RANDOM_DATA.cpr(), false)
            .add(effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated).build())
            .add(effectuatedVaccination(vaccineId, DRUG_ID, UUID.randomUUID().toString(), UUID.randomUUID().toString(), effectuated).build())
            .build();

        VaccinationEntry actualVac = vaccineMapper.toVac(vaccination);
        assertEquals(vaccination.getLatestEffectuated().effectuated().toLocalDate(), actualVac.getDt());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).ma, actualVac.getMa());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).vp, actualVac.getVp());
        assertEquals(vaccineMappingConfiguration.getVaccineIdValues(1L).mp, actualVac.getMp());
        assertEquals("DK", actualVac.getCo());
        assertEquals(VaccineMapper.SNOMED_CT_COVID_CONCEPT_ID, actualVac.getTg());
        assertEquals(2, actualVac.getDn());
        assertEquals(3, actualVac.getSd());
    }


    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());

}
