/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import dk.sds.nsp.kafka.clients.producer.NSPKafkaProducer;
import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType;
import dk.sundhedsdatastyrelsen.cbs.rules.VaccinationPassportEvaluator;
import dk.sundhedsdatastyrelsen.cbs.service.Service;
import dk.sundhedsdatastyrelsen.cbs.service.security.Security;
import dk.sundhedsdatastyrelsen.cbs.service.security.SignerFactory;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.PassportFactory;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.VaccineMapper;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.VaccineMappingConfiguration;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.io.Resource;
import org.springframework.lang.NonNull;
import org.yaml.snakeyaml.Yaml;
import se.digg.dgc.service.impl.DefaultDGCEncoder;
import se.digg.dgc.service.impl.DefaultDKEncoder;

import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.time.Duration;
import java.time.temporal.ChronoUnit;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;

@Configuration
@PropertySource("classpath:unittest.properties")
public class TestConfiguration {

    @Primary
    @Bean
    public Service service() {
        Service mockService = mock(Service.class);
        Mockito.lenient().when(mockService.createDataFetcherJob(anyString(), anyString(), any(JobType.class))).thenAnswer(i -> UUID.randomUUID().toString());
        return mockService;
    }

    @Bean
    public VaccineMapper vaccineMapper(@Value("${vaccinemapping.path:vaccinemapping.yml}") Resource mapping) {

        try {
            HashMap<Long, HashMap<String, String>> raw = new Yaml().load(mapping.getInputStream());
            Map<Long, VaccineMappingConfiguration.VaccineInformation> mappingTable = raw.entrySet().stream().collect(Collectors.toMap(Map.Entry::getKey, f -> new VaccineMappingConfiguration.VaccineInformation(f.getValue().get("vp"), f.getValue().get("mp"), f.getValue().get("ma"))));
            return new VaccineMapper(new VaccineMappingConfiguration(mappingTable));
        } catch (IOException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    @Bean
    public VaccinationPassportEvaluator vaccinationValidator() {
        return new VaccinationPassportEvaluator(Duration.of(14 * 24, ChronoUnit.HOURS), true, Duration.of(180 * 24, ChronoUnit.HOURS), true, Duration.of(14 * 24, ChronoUnit.HOURS), true);
    }

    @Bean
    public PassportFactory passportFactory(VaccineMapper vaccineMapper, VaccinationPassportEvaluator vaccinationValidator) {
        try {
            KeyStore keystore = KeyStore.getInstance("PKCS12");
            keystore.load(this.getClass().getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());

            SignerFactory signerFactory = new SignerFactory(keystore, "covid-19 borgerstatus (funktionscertifikat)", "Test1234".toCharArray(), null);
            DefaultDGCEncoder euEncoder = new DefaultDGCEncoder(signerFactory.dgcSigner());
            euEncoder.setTransliterateNames(true);
            DefaultDKEncoder dkEncoder = new DefaultDKEncoder(signerFactory.dkSigner());
            return new PassportFactory(euEncoder, dkEncoder, vaccineMapper, vaccinationValidator);
        } catch (KeyStoreException | IOException | NoSuchAlgorithmException | CertificateException | UnrecoverableKeyException e) {
            throw new RuntimeException("Load of keystore failed:" + e.getMessage(), e);
        }
    }

    @Bean
    public Security mockSecurity() {
        return new Security() {
            @NonNull
            @Override
            public String checkSecurityTicketGetCpr(@NonNull SecurityContext securityContext, @NonNull GetCoronaPassRequest request) {
                return request.getIdentifier();
            }
        };
    }

    @Bean
    public Function<Properties, NSPKafkaProducer<String, DataFetcherJob>> nspKafkaProducerConstructor() {
        return Mockito.mock(Function.class);
    }
}
