/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.Supplier;

import javax.xml.bind.JAXB;
import javax.xml.bind.JAXBException;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.TestInstance.Lifecycle;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.w3c.dom.Document;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.LibertyResponse;
import dk.sosi.seal.vault.CredentialVault;
import dk.sundhedsdatastyrelsen.cbs.configuration.BorgerContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.configuration.RequestContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.XmlDocumentResponseInterceptor;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver;
import dk.sundhedsdatastyrelsen.cbs.ddvdriver.DDVDriver.Citizen;
import dk.sundhedsdatastyrelsen.cbs.nasdriver.NASDriver;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.cbs.webservice.idws.IdwsUser;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.CoronaPassPort;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;

@ExtendWith(SpringExtension.class)
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
@TestInstance(Lifecycle.PER_CLASS)
class CoronaPassITCase {
    private static final Logger log = Logger.getLogger(CoronaPassITCase.class);
    private static final String FMK_AUDIENCE = "https://fmk";
    private static final String IDWS_CPR = "3108732076";
    private static final String IDWS_PID = "PID:9208-2002-2-098076399641";
    static final long PFIZER_VACCINE_ID = 32115515736L;
    static final long PFIZER_DRUG_ID = 32115514735L;
    URL dataFetchEndpoint = null;
    URL dataInvalidateEndpoint = null;
    
    
    private String REQUEST_CPR = "3108732076";
    private String WS_CPR = "0405870102";

    @Value("${data.fetch.url:#{null}}")
    String dataFetchUrl;
    
    @Value("${data.invalidate.url:#{null}}")
    String dataInvalidateUrl;

    @Autowired
    CoronaPassPort servicePort;
    @Autowired
    BorgerContext borgerContext;
    @Autowired
    RequestContext requestContext;
    @Autowired
    XmlDocumentResponseInterceptor documentInterceptor;
    @Autowired
    CredentialVault trustVault;
    @Autowired
    OIOIDWSFactory oioIdwsFactory;
    @Autowired
    DDVDriver ddvDriver;
    @Autowired
    NASDriver nasDriver;


    @BeforeAll
    public void setupAll() throws JAXBException {
        nasDriver.setUp();
    }
    
    @AfterAll
    public void tearDownAll() throws Exception {
        nasDriver.tearDown();
    }
    
    @BeforeEach
    public void setup() throws MalformedURLException {
        borgerContext.clearContext();
        requestContext.clearContext();
        
        if (dataFetchUrl != null) {
            dataFetchEndpoint = new URL(dataFetchUrl);
        }
        
        if (dataInvalidateUrl != null) {
            dataInvalidateEndpoint = new URL(dataInvalidateUrl);
        }
        
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_same_cpr_x_times() {
        for (int i = 0; i < 10; i++) {
            testGetCoronaPasses();
        }
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_different_cpr_x_times() {
        RandomData randomData = new RandomData(System.currentTimeMillis());
        for (int i = 0; i < 2; i++) {
            REQUEST_CPR = randomData.cpr();
            testGetCoronaPasses();
        }
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_all_cpr() {
        for (IdwsUser user : IdwsUser.values()) {
            getCoronaPass(user);
        }
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses() {
        /*
          DDV test1
          2401010010: Pas ej OK {"vaccinationStatus":"DELVIST_VACCINERET","documentId":"N/A","personInfo":{"personNameStructure":{"personGivenName":"Trine","personMiddleName":null,"personSurnameName":"Jørgensen"},"personCivilRegistrationIdentifier":"2401010010"},"vaccinationAgainst":"SARS-CoV-2 (COVID-19)","effectuation":[{"drugName":"Pfizer BioNTech Covid-19 vacc","effectuatedDateTime":"2021-01-24T23:00:00.000+0000","vaccine":{"vaccineIdentifier":32115515736,"vaccineName":"Pfizer BioNTech Covid-19 vacc","disease":[{"diseaseIdentifier":32115515735,"diseaseName":"Covid-19","diseaseNameDK":"Covid-19 sygdom","atc":{"code":"J07BX","text":"Andre virale vacciner"}}],"atc":{"code":"J07BX","text":"Andre virale vacciner"},"vaccineKeywordsText":"","displayMinimumVaccinationPlan":[],"ssidrug":[]},"ssidrug":{"atc":{"code":null,"text":null},"drugIdentifier":32115514735,"drugName":"Pfizer BioNTech COVID-19 Vacc","drugForm":{"drugFormCode":null,"drugFormText":"injektionsvæske, suspension"},"drugStrength":{"content":[{"name":"{http://vaccinationsregister.dk/schemas/2013/12/01}DrugStrengthText","declaredType":"java.lang.String","scope":"dk.vaccinationsregister.schemas._2013._12._01.DrugStrengthType","value":"","nil":false,"globalScope":false,"typeSubstituted":false}]}}}],"vaccinationPassportAsPDF":null}
          2401010011: Pas ej OK, der er kun givet 1 vaccination
          2401010012: Pas OK (Moderna) {"vaccinationStatus":"VACCINERET","documentId":"1000110","personInfo":{"personNameStructure":{"personGivenName":"Ina","personMiddleName":null,"personSurnameName":"Kristensen"},"personCivilRegistrationIdentifier":"2401010012"},"vaccinationAgainst":"SARS-CoV-2 (COVID-19)","effectuation":[{"drugName":"Moderna Covid-19 Vaccine","effectuatedDateTime":"2021-03-22T08:04:00.000+0000","vaccine":{"vaccineIdentifier":32115518136,"vaccineName":"Moderna Covid-19 Vaccine","disease":[{"diseaseIdentifier":32115515735,"diseaseName":"Covid-19","diseaseNameDK":"Covid-19 sygdom","atc":{"code":"J07BX","text":"Andre virale vacciner"}}],"atc":{"code":"J07BX","text":"Andre virale vacciner"},"vaccineKeywordsText":"","displayMinimumVaccinationPlan":[],"ssidrug":[]},"ssidrug":{"atc":{"code":null,"text":null},"drugIdentifier":32115518135,"drugName":"Moderna Covid-19 vaccine","drugForm":{"drugFormCode":null,"drugFormText":"injektionsvæske, suspension"},"drugStrength":{"content":[]}}},{"drugName":"Moderna Covid-19 Vaccine","effectuatedDateTime":"2021-04-24T07:04:00.000+0000","vaccine":{"vaccineIdentifier":32115518136,"vaccineName":"Moderna Covid-19 Vaccine","disease":[{"diseaseIdentifier":32115515735,"diseaseName":"Covid-19","diseaseNameDK":"Covid-19 sygdom","atc":{"code":"J07BX","text":"Andre virale vacciner"}}],"atc":{"code":"J07BX","text":"Andre virale vacciner"},"vaccineKeywordsText":"","displayMinimumVaccinationPlan":[],"ssidrug":[]},"ssidrug":{"atc":{"code":null,"text":null},"drugIdentifier":32115518135,"drugName":"Moderna Covid-19 vaccine","drugForm":{"drugFormCode":null,"drugFormText":"injektionsvæske, suspension"},"drugStrength":{"content":[]}}}],"vaccinationPassportAsPDF":null}
          2401010013: Pas OK (Pfizer)
          2401010014: Pas ej OK (min.interval ikke overholdt)
          2401010015: Pas ej OK (kun 1 vaccination)
          2401010016: Pas OK (Pfizer)
          2401010017: Pas OK (Pfizer)
          2401010018: Pas OK (krydsforløb med AstraZeneca + Moderna)

          insert-kids-data.sql (note that '2401010010' has no data):
          (person_identifier, test_identifier, type_of_test, test_name, test_manufacturer, timestamp_sample, timestamp_result, result_of_the_test, testing_centre_or_facility, timestamp_last_update, sletteflag)
          ('2401010011', 2021001016099, 'SARS2', '', '', '2021-01-12 10:29:00', '2021-01-12 20:19:00', 'NEGATIV', '', '2021-04-22 03:51:37', 0),
          ('2401010012', 2021010110318, 'SARS2', '', '', '2021-03-11 09:15:00', '2021-04-23 10:40:00', 'NEGATIV', '', '2021-04-23 11:23:40', 0),
          ('2401010012', 2021013666631, 'SARSG', '', '', '2021-02-20 10:30:00', '2021-02-20 10:50:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010013', 2021013966213, 'SARSG', '', '', '2021-03-23 07:09:00', '2021-03-23 07:29:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010014', 2021013966258, 'SARSG', '', '', '2021-03-23 08:06:00', '2021-03-23 08:26:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010015', 2021016153907, 'SARSG', '', '', '2021-03-31 16:40:00', '2021-03-31 17:00:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010016', 2021017448428, 'SARS2', '', '', '2021-04-05 08:00:00', '2021-04-05 18:36:00', 'NEGATIV', '', '2021-04-23 12:24:43', 0),
          ('2401010017', 2021017561852, 'SARSG', '', '', '2021-04-05 17:50:00', '2021-04-05 18:10:00', 'POSITIV', 'CAPA Biotech Aps', '2021-04-23 11:27:13', 0),
          ('2401010018', 2021018005363, 'SARSG', '', '', '2021-04-06 10:55:00', '2021-04-06 11:15:00', 'NEGATIV', 'SEI kommunal- og skoleindberetninger ', '2021-04-22 14:57:34', 0),
          ('2401010018', 2021018005364, 'SARSG', '', '', '2021-04-06 10:36:00', '2021-04-06 10:56:00', 'NEGATIV', 'SEI kommunal- og skoleindberetninger ', '2021-04-22 14:57:34', 0),
         */

        getCoronaPass(IdwsUser.CPR_3108732076);
    }

    private GetCoronaPassResponse getCoronaPass(IdwsUser user) {
        String messageId = UUID.randomUUID().toString();
        requestContext.setRequestContext(messageId);
        borgerContext.setContext(FMK_AUDIENCE, user, null);
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(user.cpr());

//        ((BindingProvider) servicePort).getRequestContext().put(MessageContext.HTTP_REQUEST_HEADERS, Collections.singletonMap("Xcachehit", Collections.singletonList("false")));

        GetCoronaPassResponse response = servicePort.getCoronaPass(request);
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        JAXB.marshal(response, os);
        log.info("GetCoronaPassResponse=" + os);

        assertNotNull(response);
        Document responseDocument = documentInterceptor.getDocument();
        LibertyResponse libertyResponse = oioIdwsFactory.createResponse(responseDocument);
        assertNotNull(libertyResponse);

        // Notice: in order to validate response signature, we need to add the cert from "oio_idws_response_signing.jks" residing
        // in NSP docker image /pack/wildfly/modules/system/layers/base/dk/sds/nsp/accesshandler/main/ to our trustVault (trust.jks)
        libertyResponse.validateSignatureAndTrust(trustVault);
        assertEquals(messageId, libertyResponse.getRelatesTo());
        
        return response;
    }
    
    @Test
    void test_A1_WhenCacheIsEmptyFutureIsReturned() throws Exception {
        
        Instant now = Instant.now();

        // Clear citizen in DDV and add+remove vaccination to trigger invalidation
        Citizen citizen = ddvDriver.withCitizen(IdwsUser.CPR_3108732076);
        citizen.clear().addVaccination(PFIZER_VACCINE_ID, PFIZER_DRUG_ID, Date.from(now)).deleteVaccination();

        GetCoronaPassResponse response = waitFor("passport to become invalidated", () -> {
            triggerInvalidation();
            return getCoronaPass(IdwsUser.CPR_3108732076);
        }, r -> r.getFutureResult() != null, 10000);

        assertNull(response.getCoronaPasses());
        assertNotNull(response.getFutureResult());
        
        waitFor("passport data to become present", () -> {
            triggerDataFetch();
            return getCoronaPass(IdwsUser.CPR_3108732076);
        }, r -> r.getFutureResult() == null, 10000);
        
        log.debug("Done");
        
    }

    @Test
    void test_A1_1_WhenCacheChangesToReadyNASMessageIsSent() throws Exception {
        Instant now = Instant.now();

        // Clear citizen in DDV and add+remove vaccination to trigger invalidation
        Citizen citizen = ddvDriver.withCitizen(IdwsUser.CPR_3108732076);
        citizen.clear().addVaccination(PFIZER_VACCINE_ID, PFIZER_DRUG_ID, Date.from(now)).deleteVaccination();

        GetCoronaPassResponse response = waitFor("passport to become invalidated", () -> {
            triggerInvalidation();
            return getCoronaPass(IdwsUser.CPR_3108732076);
        }, r -> r.getFutureResult() != null, 10000);

        assertNull(response.getCoronaPasses());
        assertNotNull(response.getFutureResult());
        
        String jobId = response.getFutureResult().getJobId();
        
        waitFor("NAS message with jobId " + jobId + " to arrive", () -> {
            triggerDataFetch();
            return nasDriver.getNotifications();

        }, r -> r.stream().anyMatch(n -> n.getJobId().equals(jobId)), 10000);
        
        log.debug("Done");
    }
    

    @Disabled("NYI")
    @Test
    void test_A1_2_AfterNASMessageIsSentDataCanBeFetched() throws Exception {
        Instant now = Instant.now();

        // Clear citizen in DDV and add+remove vaccination to trigger invalidation
        Citizen citizen = ddvDriver.withCitizen(IdwsUser.CPR_3108732076);
        citizen.clear();
        citizen.addVaccination(PFIZER_VACCINE_ID, PFIZER_DRUG_ID, Date.from(now));
        citizen.addVaccination(PFIZER_VACCINE_ID, PFIZER_DRUG_ID, Date.from(now.minus(28, ChronoUnit.DAYS)));

        
        GetCoronaPassResponse response = waitFor("passport to become invalidated", () -> {
            triggerInvalidation();
            return getCoronaPass(IdwsUser.CPR_3108732076);
        }, r -> r.getFutureResult() != null, 100000);

        assertNull(response.getCoronaPasses());
        assertNotNull(response.getFutureResult());
        
        waitFor("NAS message to arrive", () -> {
            triggerDataFetch();
            return nasDriver.getMessages();

        }, m -> m.size() > 0, 10000);

        GetCoronaPassResponse passport = getCoronaPass(IdwsUser.CPR_3108732076);
        
        Assertions.assertFalse(passport.getCoronaPasses().getPass().isEmpty());
    }

    @Disabled("NYI")
    @Test
    void test_A2_PassportIsReturnedWithFreshSignature() throws Exception {
        fail();
    }

    @Disabled("NYI")
    @Test
    void test_A3_OnInvalidDataNASMessageIsSent() throws Exception {
        fail();
    }
    
    @Disabled("NYI")
    @Test
    void test_A5_PassportWithOnlyTestResultIsReturnedWithFreshSignature() throws Exception {
        fail();
    }
    
    @Disabled("NYI")
    @Test
    void test_A6_PassportWithOnlyVaccinationIsReturnedWithFreshSignature() throws Exception {
        fail();
    }

    @Disabled("NYI")
    @Test
    void test_A7_PassportWithOnlyImmunityIsReturnedWithFreshSignature() throws Exception {
        fail();
    }

    @Disabled("NYI")
    @Test
    void test_A8_PassportWithNoResultIsReturnedWithFreshSignature() throws Exception {
        fail();
    }

    @Disabled("NYI")
    @Test
    void test_A9_PassportWithNoDataIsReturnedWithFreshSignature() throws Exception {
        fail();
    }

    @Test 
    void test_A10_CPRMismatch() throws Exception {
        IdwsUser user = IdwsUser.CPR_0405870101;
        
        String messageId = UUID.randomUUID().toString();
        requestContext.setRequestContext(messageId);
        borgerContext.setContext(FMK_AUDIENCE, user, null);
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        
        // Create cpr mismatch: 
        request.setIdentifier("1111111118");
        
        Assertions.assertThrows(SOAPFaultException.class, () -> {
            servicePort.getCoronaPass(request); 
        });
        
    }
    
    @Disabled("NYI")
    @Test 
    void test_A11_PersonNotInCPR() throws Exception {
        fail();
    }
    
    @Disabled("NYI")
    @Test
    void test_A12_GeneralDataFetchError() throws Exception {
        fail();
    }
    
    @Disabled("NYI")
    @Test
    void test_A13_GeneralTechnicalError() throws Exception {
        fail();
    }
    
    private void triggerDataFetch() {
        if (dataFetchEndpoint == null) {
            log.debug("Trigger data fetch disabled");
            return;
        }
        
        try (InputStream is = dataFetchEndpoint.openStream()) {
            log.debug("Data fetch responded with: " + IOUtils.toString(is, StandardCharsets.ISO_8859_1));
        } catch (IOException e) {
            log.warn("Trigger data fetch failed: " + e.getMessage());
        }
    }

    private void triggerInvalidation() {
        if (dataInvalidateEndpoint == null) {
            log.debug("Trigger data invalidate disabled");
            return;
        }
        
        try (InputStream is = dataInvalidateEndpoint.openStream()) {
            log.debug("Data invalidate responded with: " + IOUtils.toString(is, StandardCharsets.ISO_8859_1));
        } catch (IOException e) {
            log.warn("Trigger data invalidate failed: " + e.getMessage());
        }
    }

    @SuppressWarnings("java:S2925")
    private <T> T waitFor(String what, Supplier<T> supplier, Function<T, Boolean> evaluator, int maxWaitMillis) throws InterruptedException {
        Instant start = Instant.now();

        while (start.plus(maxWaitMillis, ChronoUnit.MILLIS).isAfter(Instant.now())) {
            log.debug("Checking for " + what);
            T result = supplier.get();
            if (evaluator.apply(result)) {
                log.debug("Got expected result");
                return result;
            }
            Thread.sleep(1000);
        }

        fail("Timeout waiting for " + what);

        return null;
    }

}
