/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nasdriver;

import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Holder;
import javax.xml.ws.wsaddressing.W3CEndpointReference;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wsn.b_2.DestroyPullPoint;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.Unsubscribe;
import org.oasis_open.docs.wsn.bw_2.CreatePullPoint;
import org.oasis_open.docs.wsn.bw_2.NotificationBrokerService;
import org.oasis_open.docs.wsn.bw_2.NotificationConsumer;
import org.oasis_open.docs.wsn.bw_2.NotificationProducer;
import org.oasis_open.docs.wsn.bw_2.PullPointFactoryService;
import org.oasis_open.docs.wsn.bw_2.SubscriptionManagerService;
import org.w3c.dom.Node;

import com.sun.xml.ws.dump.MessageDumpingFeature;

import dk.nsi.advis.v10.NotifyContent;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.DGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.DGWSHeaders;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.DefaultDGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.EndpointReferences;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClientImpl;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.PullPointFactory;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatusNotification;

public class NASDriver {
    private static final Logger logger = Logger.getLogger(NASDriver.class);

    public static final String JOB_COMPLETED_TOPIC = "http://nspop.dk/GreenCertificate/2021/04/01:JobCompleted";
    private static JAXBContext jaxbContext; 
    static {
        try {
            jaxbContext = JAXBContext.newInstance(JobStatusNotification.class);
        } catch (JAXBException e) {
            logger.error("Unable to create JAXBContext for JobStatusNotification");
        }
    }
    private final org.oasis_open.docs.wsn.b_2.ObjectFactory b2ObjectFactory = new org.oasis_open.docs.wsn.b_2.ObjectFactory();

    private DGWSHeaderProvider headerProvider;
    private PullPointFactory pullPointFactory;
    private W3CEndpointReference pullPoint;
    private W3CEndpointReference subscribe;
    private NasClientImpl nasClient;
    private NotificationProducer notificationProducer;


    
    private final MessageDumpingFeature messageDumpingFeature = new MessageDumpingFeature(true, NASDriver.class.getName(), Level.INFO.toString(), true);

    private String stsUrl;
    private String notificationBrokerUrl;
    private String pullPointFactoryUrl;
    private String subscriptionManagerUrl;

    public NASDriver(String stsUrl, String notificationBrokerUrl, String pullPointFactoryUrl, String subscriptionManagerUrl) {
        
        this.stsUrl = stsUrl;
        this.notificationBrokerUrl = notificationBrokerUrl;
        this.pullPointFactoryUrl = pullPointFactoryUrl;
        this.subscriptionManagerUrl = subscriptionManagerUrl;

        System.setProperty(DGWSRequestFactory.class.getName() + ".test", "true");
    }
    
    public void setUp() throws JAXBException {
        messageDumpingFeature.disableMessageLogging();
        STSConfig stsConfig = new STSConfig(
            stsUrl,
            "Sundhedsdatastyrelsen",
            "33257872",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        NotificationConsumer soap = new NotificationBrokerService().getSoap();
        setEndpoint((BindingProvider) soap, notificationBrokerUrl);

        headerProvider = new DefaultDGWSHeaderProvider(new DGWSRequestFactory(stsConfig, new dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl(new HttpExecutor())));
        pullPointFactory = new PullPointFactory();

        notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, subscriptionManagerUrl);

        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, pullPointFactoryUrl);

        nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);

        pullPoint = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);
    }

    public void tearDown() throws Exception {
        DGWSHeaders headers = headerProvider.getHeaders();

        if (subscribe != null) {
            String subscribeAddress = EndpointReferences.extractAddress(subscribe);
            NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
            setEndpoint((BindingProvider) notificationProducer, subscribeAddress);
            notificationProducer.unsubscribe(new Unsubscribe(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        }

        if (pullPoint != null) {
            pullPointFactory.getPullPoint(pullPoint).destroyPullPoint(new DestroyPullPoint(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        }
    }
    
    public List<NotificationMessageHolderType> getMessages() {

        return nasClient.getMessages(pullPoint, 1000, UUID.randomUUID());
    }

    public List<JobStatusNotification> getNotifications() {
        
        List<NotificationMessageHolderType> messages = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID());
        
//        messages.stream().map(n -> )

        return messages.stream().map(mes -> unmarshal(mes)).collect(Collectors.toList());
    }

    
    
    private JobStatusNotification unmarshal(NotificationMessageHolderType mes) {
        try {
            NotifyContent notifyContent = (NotifyContent)mes.getMessage().getAny();

            Object any = notifyContent.getAny();
            
            System.out.println("any: " + any + ": " + any.getClass());
            
            Unmarshaller jaxbUnmarshaller = jaxbContext.createUnmarshaller();

            JAXBElement<JobStatusNotification> jaxbElement = (JAXBElement<JobStatusNotification>) jaxbUnmarshaller.unmarshal((Node)any, JobStatusNotification.class);
            
            return jaxbElement.getValue();

        } catch (JAXBException e) {
            throw new RuntimeException("Unable to unmarshal", e);
        }

    }

    private void setEndpoint(BindingProvider bindingProvider, String url) {
        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, url);
    }


}
